"""
Utils for the Director Modules

These directors correspond to the PymodaqListener
"""

from typing import Optional, Union, List

from pyleco.directors.director import Director

import pymodaq_gui.parameter.utils as putils
from pymodaq_gui.parameter import Parameter, ioxml
from pymodaq.utils.data import DataActuator
from pymodaq.utils.leco.utils import binary_serialization_to_kwargs, SerializableFactory

from pymodaq_gui.parameter.utils import ParameterWithPath



class GenericDirector(Director):
    """Director helper to control some Module remotely."""

    def set_remote_name(self, name: Optional[str] = None):
        """Set the remote name of the Module (i.e. where it should send responses to)."""
        self.ask_rpc(method="set_remote_name", name=name or self.communicator.name)

    def set_info(self, param: Parameter):
        # It removes the first two parts (main_settings and detector_settings?)
        pwp = ParameterWithPath(param, putils.get_param_path(param)[2:])
        self.ask_rpc(method="set_info",
                     **binary_serialization_to_kwargs(pwp, data_key='parameter'))

    def get_settings(self,) -> None:
        self.ask_rpc('get_settings')


class DetectorDirector(GenericDirector):
    def send_data_grab(self, grabber_type: str = "") -> None:
        self.ask_rpc("send_data_grab", grabber_type=grabber_type)

    def send_data_snap(self, grabber_type: str = "") -> None:
        self.ask_rpc("send_data_snap", grabber_type=grabber_type)

    def stop_grab(self, grabber_type: str = "") -> None:
        self.ask_rpc("stop_grab", grabber_type=grabber_type)


class ActuatorDirector(GenericDirector):
    def move_abs(self, position: Union[float, DataActuator]) -> None:
        self.ask_rpc(
            "move_abs", **binary_serialization_to_kwargs(position, data_key="position")
        )

    def move_rel(self, position: Union[float, DataActuator]) -> None:
        self.ask_rpc(
            "move_rel", **binary_serialization_to_kwargs(position, data_key="position")
        )

    def move_home(self) -> None:
        self.ask_rpc("move_home")

    def get_actuator_value(self) -> None:
        """Request that the actuator value is sent later on.

        Later the `set_data` method will be called.
        """
        # according to DAQ_Move, this supersedes "check_position"
        self.ask_rpc("get_actuator_value")

    def stop_motion(self,) -> None:
        # not implemented in DAQ_Move!
        self.ask_rpc("stop_motion")

